from yaqq_ds import GenerateDataSet, VisualizeDataSet, ResultsPlotSave
from yaqq_nus import NovelUniversalitySearch

def run():

    print("\n_____________________________________________________________________")
    print("\n           Welcome to YAQQ: Yet Another Quantum Quantizer.           ")
    print("\nCopyright © 2023 Quantum Intelligence Research Group; AGPL v3 License")
    print("\n  Code repository: https://github.com/Advanced-Research-Centre/YAQQ  ")
    print("_____________________________________________________________________")

    gds = GenerateDataSet()
    vds = VisualizeDataSet()
    nsa = NovelUniversalitySearch()
    rps = ResultsPlotSave()

    devmode = input("\n  ===> Run Default Configuration? [Y/N] (def.: Y): ") or 'Y'

    if devmode == 'Y':
        
        # [ 1q:'rand','skt' | 2q:'rand','kak' | 3+q:'rand','qsd' ]
        # All possible options for testing: 
        #   3+q: [x,x,1], [1,1,2], [1,2,2], [2,1,2], [2,2,2]
        #    2q: [x,1,-], [1,2,-], [2,2,-]
        #    1q: [1,-,-], [2,-,-]
        # Most analytical, best fidelity, slow = [2,2,2] : 'skt','kak','qsd'   
        # Least analytical, bad fidelity, fast = [1,1,1] : 'rand','rand','rand'
        yaqq_cf_dcmp = [2,2,2]
        nsa.cnfg_dcmp(yaqq_cf_dcmp)

        # nsa.decompose_u()                   
         
        yaqq_ds_dim = 3
        yaqq_ds_type = 2
        yaqq_ds_size = 5
        yaqq_ds_reso = 23
        yaqq_ds = gds.yaqq_gen_ds(yaqq_ds_dim, yaqq_ds_type, yaqq_ds_size, yaqq_ds_reso)
            
        # gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db = nsa.compare_gs(yaqq_ds)
        # rps.plot_compare_gs(gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, pfivt = True) 

        yaqq_cf_wgts = [100,1,50,1,0]
        nsa.cnfg_wgts(yaqq_cf_wgts)
        # yaqq_cf_ngs = ['R1','R1','CX2']
        # yaqq_cf_ngs = ['H1','T1','SPE2']
        yaqq_cf_ngs = ['H1','P1','CX2']
        yaqq_ngs_search = 2

        gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, opt_params = nsa.nusa(yaqq_ds,yaqq_cf_ngs,yaqq_ngs_search)
        for i in gs2:
            print(i, gs2[i])
        print(opt_params)
        rps.plot_compare_gs(gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, pfivt = True) 

    elif devmode == 'N':

        print("\n  YAQQ has 4 configuration options:")
        print("     1. Operation Mode")
        print("     2. Gate Decomposition Method")
        print("     3. Cost Function")
        print("     4. Data Dimension, Type and Size")

        # TBD: Checks for invalid configurations
        print("\n Operation Mode:")
        print("   1. Generative novel GS2 w.r.t. GS1 (in code)")
        print("   2. Compare GS2 (in code) w.r.t. GS1 (in code)")
        print("   3. Decompose a n-qubit U (in code) w.r.t. GS1 (in code)")
        yaqq_mode = int(input("\n  ===> Enter YAQQ Mode (def.: 1): ") or 1)

        yaqq_cf_dcmp = []
        print("\n Gate Decomposition Method for Dimension = 1:")
        print("     1. Random Decomposition (trails auto-scale w.r.t. dimension)")
        print("     2. Solovay-Kitaev Decomposition")
        yaqq_cf_dcmp.append(int(input("\n  ===> Enter Gate Decomposition Method for Dimension = 2 (def.: 2): ") or 2))
        print("\n Gate Decomposition Method for Dimension = 2:")
        print("     1. Random Decomposition (trails auto-scale w.r.t. dimension)")
        print("     2. Cartan Decomposition")
        yaqq_cf_dcmp.append(int(input("\n  ===> Enter Gate Decomposition Method for Dimension = 2 (def.: 2): ") or 2))
        print("\n Gate Decomposition Method for Dimension = 3+:")
        print("     1. Random Decomposition (trails auto-scale w.r.t. dimension)")
        print("     2. Quantum Shannon Decomposition")
        yaqq_cf_dcmp.append(int(input("\n  ===> Enter Gate Decomposition Method for Dimension = 3+ (def.: 2): ") or 2))
        nsa.cnfg_dcmp(yaqq_cf_dcmp)

        if yaqq_mode == 3:
            nsa.decompose_u()    # Dataset and costfunction selection not required for this mode
            print("\n_____________________________________________________________________")
            print("\n--------------------- Thank you for using YAQQ. ---------------------")
            print("_____________________________________________________________________")
            exit()

        yaqq_ds_dim = int(input("\n  ===> Enter Data Set Dimension (def.: 2): ") or 2)        
        if yaqq_ds_dim == 1:
            print("\n Data Set Types for Dimension = 1:")
            print("     1. Haar Random 1-qubit pure States")
            print("     2. Haar Random 2x2 Unitaries")
            print("     3. Equispaced States on Bloch Sphere using Golden mean")
            print("     4. Equispaced Angles on Bloch Sphere")
            yaqq_ds_type = int(input("\n  ===> Enter Data Set Type (def.: 3): ") or 3)
        elif yaqq_ds_dim == 2:
            print("\n Data Set Types for Dimension = 2:")
            print("     1. Haar Random 2-qubit pure States")
            print("     2. Haar Random 4x4 Unitaries")
            print("     3. Random Non-local Unitaries on Weyl chamber")
            print("     4. Equispaced Non-local Unitaries on Weyl chamber")
            yaqq_ds_type = int(input("\n  ===> Enter Data Set Type (def.: 4): ") or 4)
        else:
            print("\n Data Set Types for Dimension = 3+:")
            print("     1. Haar Random n-qubit pure States")
            print("     2. Haar Random 2^nx2^n Unitaries")
            yaqq_ds_type = int(input("\n  ===> Enter Data Set Type (def.: 2): ") or 2)

        if yaqq_ds_dim == 1 and yaqq_ds_type == 4:
            yaqq_ds_reso = int(input("\n  ===> Enter Bloch Sphere a_rz Spacing (def.: 16, 512 points): ") or 16)
            yaqq_ds = gds.yaqq_gen_ds(yaqq_ds_dim, yaqq_ds_type, None, yaqq_ds_reso)
        elif yaqq_ds_dim == 2 and yaqq_ds_type == 4:
            yaqq_ds_reso = int(input("\n  ===> Enter Weyl Chamber cx Spacing (def.: 23, 508 points): ") or 23)
            yaqq_ds = gds.yaqq_gen_ds(yaqq_ds_dim, yaqq_ds_type, None, yaqq_ds_reso)
        else:
            yaqq_ds_size = int(input("\n  ===> Enter Data Set Size (def.: 508): ") or 508)
            yaqq_ds = gds.yaqq_gen_ds(yaqq_ds_dim, yaqq_ds_type, yaqq_ds_size, None)

        if yaqq_ds_dim <= 2:
            yaqq_ds_show = input("\n  ===> Visualize Data Set? [Y/N] (def.: N): ") or 'N'
            if yaqq_ds_show == 'Y':
                if yaqq_ds_dim == 1:
                    vds.vis_ds_Bloch(yaqq_ds)
                else:
                    vds.vis_ds_Weyl(yaqq_ds)

        if yaqq_mode == 2:
            gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db = nsa.compare_gs(yaqq_ds)     # Costfunction selection not required for this mode
            rps.plot_compare_gs(gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, pfivt = True) 
            print("\n_____________________________________________________________________")
            print("\n--------------------- Thank you for using YAQQ. ---------------------")
            print("_____________________________________________________________________")
            exit()

        yaqq_cf_wgts = [int(i) for i in (input("\n  ===> Enter Cost Function Weights (def.: [1,1,1,1,0]): ") or '1,1,1,1,0').split(',')]
        print("\n  ===> Cost Function Weights: ", yaqq_cf_wgts)
        nsa.cnfg_wgts(yaqq_cf_wgts)

        print("\n Novel Gate Set Composition:")
        print("   R1: Haar Random 1-qubit Unitary")                         # Search: random
        print("   P1: Parametric 1-qubit Unitary (IBM U3)")                 # Search: parametric, random
        print("   G1: Golden 1-qubit Unitary")                              # TBD
        print("   SG1: Super Golden 1-qubit Unitary")                       # TBD
        print("   T1: T Gate 1-qubit Unitary")                              # Constant
        print("   TD1: T-dagger Gate 1-qubit Unitary")                      # Constant
        print("   H1: H (Hadamard) Gate 1-qubit Unitary")                   # Constant
        print("   F1: Load 1-qubit Unitary Gate definition from File")      # Constant TBD Extension
        if yaqq_ds_dim >= 2:
            print("   R2: Haar Random 2-qubit Unitary")                     # Search: random
            print("   NL2: Non-local 2-qubit Unitary")                      # Search: parametric, random
            print("   CX2: CNOT Gate 2-qubit Unitary")                      # Constant
            print("   B2: B (Berkeley) Gate 2-qubit Unitary")               # Constant
            print("   PE2: Perfect Entangler 2-qubit Unitary")              # TBD
            print("   SPE2: Special Perfect Entangler 2-qubit Unitary")     # Search: parametric, random
            print("   F2: Load 2-qubit Unitary Gate definition from File")  # Constant TBD Extension
        yaqq_cf_ngs = (input("\n  ===> Enter Get Set (def.: [R1,R1,R1]): ") or 'R1,R1,R1').split(',')

        print("\n Search Method:")
        print("   1: Random Search for non-constant gates")
        print("   2: Parametric Search (SciPy) for non-constant gates")
        yaqq_ngs_search = int(input("\n  ===> Enter Search Method (def.: 1): ") or 1)

        gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, opt_params = nsa.nusa(yaqq_ds,yaqq_cf_ngs,yaqq_ngs_search)
        print("Novel Parameters for GS2: ", opt_params)
        rps.plot_compare_gs(gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db, pfivt = True)  
        print("\n_____________________________________________________________________")
        print("\n--------------------- Thank you for using YAQQ. ---------------------")
        print("_____________________________________________________________________")

########################################################################################################################################################################################################


import numpy as np
from qiskit import QuantumCircuit
from qiskit.quantum_info import random_unitary, Choi, process_fidelity, Operator
from qiskit.extensions import UnitaryGate
from qiskit.quantum_info.synthesis import qsd
import random
from skt import gen_basis_seq, UGate, UdgGate
from qiskit.transpiler.passes.synthesis import SolovayKitaev
from tqdm import tqdm
import time
import weylchamber
from scipy.optimize import minimize 
from qiskit.quantum_info import TwoQubitBasisDecomposer
from qiskit.quantum_info.synthesis.two_qubit_decompose import TwoQubitWeylDecomposition
import warnings

class NovelUniversalitySearch:

    # ------------------------------------------------------------------------------------------------ #

    """
    Configure the decomposition methods for 1, 2 and 3+ qubit gates
    """

    def cnfg_dcmp(self, dcmp):

        self.d_1q = dcmp[0]    # Decomposition method for 1-qubit gates             [rand, skd]
        self.d_2q = dcmp[1]    # Decomposition method for 2-qubit gates             [rand, cartan]
        self.d_nq = dcmp[2]    # Decomposition method for 3 or more qubit gates     [rand, qsd]

        return
      
    # ------------------------------------------------------------------------------------------------ #

    def def_gs(self, gs_cfg, params = None):

        # All params are normalized to [0,1]
        gs = {}
        gno = 0
        param_ctr = 0
        for g in gs_cfg:
            gno += 1
            match g:
                case 'R1':      # R1: Haar Random 1-qubit Unitary
                    U = random_unitary(2).data
                case 'P1':      # P1: Parametric 1-qubit Unitary (Qiskit U3)
                    theta, phi, lam = params[param_ctr]*np.pi, params[param_ctr+1]*np.pi, params[param_ctr+2]*np.pi
                    U = np.asarray( [[np.cos(theta/2), -np.exp(1j*lam)*np.sin(theta/2)],
                                    [np.exp(1j*phi)*np.sin(theta/2), np.exp(1j*(lam+phi))*np.cos(theta/2)]])
                    param_ctr += 3
                case 'G1':      # G1: Golden 1-qubit Unitary
                    U = random_unitary(2).data  # TBD Extension
                case 'SG1':     # SG1: Super Golden 1-qubit Unitary
                    # Ref: https://arxiv.org/abs/1704.02106
                    U = random_unitary(2).data  # TBD Extension
                case 'T1':      # T1: T Gate 1-qubit Unitary
                    U = np.array([[1, 0], [0, (1+1j)/np.sqrt(2)]], dtype=complex)
                case 'TD1':     # TD1: T-dagger Gate 1-qubit Unitary
                    U = np.array([[1, 0], [0, (1-1j)/np.sqrt(2)]], dtype=complex)
                case 'H1':      # H1: H (Hadamard) Gate 1-qubit Unitary
                    U = np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)
                case 'F1':      # F1: Load 1-qubit Unitary Gate definition from File
                    # ask user for file name
                    pass        # TBD Extension
                case 'R2':      # R2: Haar Random 2-qubit Unitary
                    U = random_unitary(4).data
                case 'NL2':     # NL2: Non-local 2-qubit Unitary
                    U = weylchamber.canonical_gate(params[param_ctr],params[param_ctr+1],params[param_ctr+2])
                    param_ctr += 3
                case 'CX2':     # CX2: CNOT Gate 2-qubit Unitary
                    U = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]], dtype=complex)
                case 'B2':      # B2: B (Berkeley) Gate 2-qubit Unitary
                    U = weylchamber.canonical_gate(0.5,0.25,0)
                case 'PE2':     # PE2: Perfect Entangler 2-qubit Unitary
                    U = random_unitary(2).data  # TBD Extension
                case 'SPE2':    # SPE2: Special Perfect Entangler 2-qubit Unitary
                    U = weylchamber.canonical_gate(0.5,params[param_ctr],0)
                    param_ctr += 1
                case 'F2':      # F2: Load 2-qubit Unitary Gate definition from File
                    # ask user for file name
                    pass        # TBD Extension
                case _:
                    print("Invalid Gate Set Configuration")
                    exit()
            gs[str(gno)+g] = UnitaryGate(U, label=str(gno)+g) 
        gs_gates = ','.join(list(gs.keys()))

        return gs, gs_gates
    # ------------------------------------------------------------------------------------------------ #

    def gs_param_ctr(self, gs_cfg):

        param_ctr = 0
        for g in gs_cfg:
            match g:
                case 'P1':      # P1: Parametric 1-qubit Unitary (Qiskit U3)
                    param_ctr += 3
                case 'G1':      # G1: Golden 1-qubit Unitary
                    param_ctr += 0  # TBD Extension
                case 'SG1':     # SG1: Super Golden 1-qubit Unitary
                    param_ctr += 0  # TBD Extension
                case 'NL2':     # NL2: Non-local 2-qubit Unitary
                    param_ctr += 3
                case 'PE2':     # PE2: Perfect Entangler 2-qubit Unitary
                    param_ctr += 0  # TBD Extension
                case 'SPE2':    # SPE2: Special Perfect Entangler 2-qubit Unitary
                    param_ctr += 1

        return param_ctr
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Given an unitary and a gate set, use trials to find a circuit using the gate set that is close to the unitary
    """

    def dcmp_rand(self, U, gs, trials, max_depth):

        choi0 = Choi(U)
        dim = int(np.log2(choi0.dim[0]))
        pfi_best = 0
        dep_best = 0
        qcirc_best = []
        for i in range(trials):
            dep = random.randrange(1,max_depth)
            seq = random.choices(list(gs.keys()), k=dep)
            qc0 = QuantumCircuit(dim)
            for seqid in seq:
                if gs[seqid].num_qubits <= dim:
                    g_order = list(range(dim))
                    random.shuffle(g_order)
                    qc0.append(gs[seqid], g_order[:gs[seqid].num_qubits])
            choi01 = Choi(qc0)
            pfi = process_fidelity(choi0,choi01)
            if pfi > pfi_best:   # TBD: Should we consider the circuit with the highest fidelity or the one with the lowest depth?
                pfi_best = pfi
                qcirc_best = qc0
                dep_best = dep

        return pfi_best, dep_best, qcirc_best

    # ------------------------------------------------------------------------------------------------ #

    """
    Convert gate set from Unitary Gate to UGate for Solovay-Kitaev decomposition of Qiskit
    """

    def skt_gs(self, gs):

        gs_skt = []
        for g in gs.keys():
            if gs[g].num_qubits == 1:
                gs_skt.append(UGate(g,gs[g].to_matrix()))

        return gs_skt
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Given an unitary and a gate set, use Qiskit's Solovay-Kitaev decomposition to find a circuit using the gate set that is close to the unitary
    """

    def dcmp_skt(self, U, skt_obj):

        choi0 = Choi(U)
        dim = int(np.log2(choi0.dim[0]))
        qc0 = QuantumCircuit(dim)
        qc0.append(U,list(range(dim)))
        qc01 = skt_obj(qc0)
        # choi0 = Choi(qc0)
        choi01 = Choi(qc01)
        pfi = process_fidelity(choi0,choi01)
        dep = qc01.depth()

        return pfi, dep, qc01
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Find a basis gate to use from the gate set for KAK decomposition
    """

    def kak_gs(self, gs):
        
        gnq = []
        for g in gs.keys():
            gnq.append(gs[g].num_qubits)

        if 2 not in gnq:
            print("No 2-qubit gate in the gate set. Cartan decomposition using 1-qubit gates not available.")
            # TBD Extension: Use 1-qubit gates to construct 2-qubit gates
            exit()
        elif gnq.count(2) > 1:
            print("Multiple choices of 2-qubit gate in the gate set. Cartan decomposition using multiple 2-qubit gates not available.")
            # TBD Extension: Use multiple 2-qubit gates
            exit()
        else:
            bg = gs[list(gs.keys())[gnq.index(2)]]

        return bg


    # ------------------------------------------------------------------------------------------------ #

    """
    Given an unitary and a gate set, decompose the unitary into the gate set based on the decomposition methods
    """

    def dcmp_U_gs(self, U, gs):
        
        dim = int(np.log2(Choi(U).dim[0]))
        rand_trials = 100
        rand_max_depth = 500

        if dim > 2 and self.d_nq == 1:      # Random n-qubit decomposition
            # Decompose using gates in gs1, no further decompositions required
            pf, cd, qc = self.dcmp_rand(U, gs, trials = rand_trials, max_depth = rand_max_depth)
        elif dim > 2 and self.d_nq == 2:    # QSD decomposition
            # Need to further decompose both 2-qubit and 1-qubit gates
            qsd_circ = qsd.qs_decomposition(U.to_matrix())
            ds_qsd_1q, ds_qsd_2q = [], []
            for g in qsd_circ:
                if g.operation.num_qubits == 1:
                    ds_qsd_1q.append(UnitaryGate(Operator(g.operation),label='QSD_1q'))
                elif g.operation.num_qubits == 2:
                    ds_qsd_2q.append(UnitaryGate(Operator(g.operation),label='QSD_2q'))
            # Decompose 2 qubits gates in ds_qsd_2q           
            ds_qsd_2q_gs = []
            for U_qsd_2q in ds_qsd_2q:
                _, _, qcirc_QSD_2q = self.dcmp_U_gs(U_qsd_2q, gs)
                ds_qsd_2q_gs.append(qcirc_QSD_2q)
            # Decompose 1 qubit gates in ds_qsd_1q
            ds_qsd_1q_gs = []
            for U_qsd_1q in ds_qsd_1q:
                _, _, qcirc_QSD_1q = self.dcmp_U_gs(U_qsd_1q, gs)
                ds_qsd_1q_gs.append(qcirc_QSD_1q)
            # Replace the gates in QSD with decomposed circuits
            qc = QuantumCircuit(dim)
            for g in qsd_circ:
                if g.operation.num_qubits == 1:
                    U_gs = ds_qsd_1q_gs.pop(0)
                    tgt = [qsd_circ.find_bit(x).index for x in g.qubits]
                    for g_gs in U_gs:
                        qc.append(g_gs.operation, tgt)
                elif g.operation.num_qubits == 2:
                    U_gs = ds_qsd_2q_gs.pop(0)
                    tgt = [qsd_circ.find_bit(x).index for x in g.qubits]
                    for g_gs in U_gs:
                        if len(g_gs.qubits) == 1:
                            qc.append(g_gs.operation, [tgt[g_gs.qubits[0].index]])
                        elif len(g_gs.qubits) == 2:
                            qc.append(g_gs.operation, tgt)
                qc.barrier()
            pf = process_fidelity(Choi(U),Choi(qc))
            cd = qc.depth()

        if dim == 2 and self.d_2q == 1:     # Random 2-qubit decomposition
            # Decompose using gates in gs1, no further decompositions required
            pf, cd, qc = self.dcmp_rand(U, gs, trials = rand_trials, max_depth = rand_max_depth)
        elif dim == 2 and self.d_2q == 2:   # Cartan decomposition
            # Analytically decompose using CAN 2-qubit gates in gs1, need to further decompose 1-qubit gates
            bg = self.kak_gs(gs)
            # warnings.filterwarnings("ignore", category=UserWarning) # Ignore warning of non-perfect decomposition for non-supercontrolled gates
            kak_obj = TwoQubitBasisDecomposer(bg)
            ds_kak_1q = kak_obj.decomp3_supercontrolled(TwoQubitWeylDecomposition(U))
            # Decompose 1 qubit gates in ds_kak_1q
            ds_kak_1q_gs = []
            for U_kak_1q in ds_kak_1q:
                _, _, qcirc_KAK_1q = self.dcmp_U_gs(UnitaryGate(Operator(U_kak_1q),label='KAK_1q'), gs)
                ds_kak_1q_gs.append(qcirc_KAK_1q)
            # Replace the gates in KAK with decomposed circuits
            qc = QuantumCircuit(2)
            for g_gs in ds_kak_1q_gs[0]:
                qc.append(g_gs.operation, [0])
            for g_gs in ds_kak_1q_gs[1]:
                qc.append(g_gs.operation, [1])
            qc.append(bg,[0,1])
            for g_gs in ds_kak_1q_gs[2]:
                qc.append(g_gs.operation, [0])
            for g_gs in ds_kak_1q_gs[3]:
                qc.append(g_gs.operation, [1])
            qc.append(bg,[0,1])
            for g_gs in ds_kak_1q_gs[4]:
                qc.append(g_gs.operation, [0])
            for g_gs in ds_kak_1q_gs[5]:
                qc.append(g_gs.operation, [1])
            qc.append(bg,[0,1])
            for g_gs in ds_kak_1q_gs[6]:
                qc.append(g_gs.operation, [0])
            for g_gs in ds_kak_1q_gs[7]:
                qc.append(g_gs.operation, [1])
            pf = process_fidelity(Choi(U),Choi(qc))
            cd = qc.depth()

        if dim == 1 and self.d_1q == 1:     # Random 1-qubit decomposition
            pf, cd, qc = self.dcmp_rand(U, gs, trials = rand_trials, max_depth = rand_max_depth)
        elif dim == 1 and self.d_1q == 2:   # Solovay-Kitaev decomposition
            gbs = gen_basis_seq()
            skt_obj = SolovayKitaev(recursion_degree = 3, basic_approximations = gbs.generate_basic_approximations(self.skt_gs(gs)))  # declare SKT object, larger recursion depth increases the accuracy and length of the decomposition
            pf, cd, qc = self.dcmp_skt(U, skt_obj)

        return pf, cd, qc

    # ------------------------------------------------------------------------------------------------ #

    def decompose_u(self):

        # Define Unitary to decompose here
        dim = 3
        U = UnitaryGate(random_unitary(2**dim),label='RndU')    # TBD: Take from user input
        
        # Define Gate Set to decompose into here
        # gs, _ = self.def_gs(['H1','T1','TD1'])
        gs, _ = self.def_gs(['H1','T1','TD1','CX2'])            # TBD: Take from user input
        # gs, _ = self.def_gs(['H1','T1','CX2'])
        
        # Decompose Unitary into Gate Set
        pf, cd, qc = self.dcmp_U_gs(U, gs)
        # print(qc)        
        print(pf, cd)
            
        return
    
    # ------------------------------------------------------------------------------------------------ #
    
    def compare_gs(self, ds):

        # Define Gate Sets here (TBD: Load saved gate set from file)

        # gs1, gs1_gates = self.def_gs(['H1','T1','TD1'])
        # gs2, gs2_gates = self.def_gs(['R1','R1','R1'])

        # gs1, gs1_gates = self.def_gs(['H1','T1','CX2'])
        # gs2, gs2_gates = self.def_gs(['R1','R1','R2']) 

        gs1, gs1_gates = self.def_gs(['H1','T1','CX2'])         # TBD: Take from user input
        gs2, gs2_gates = self.def_gs(['R1','R1','CX2'])         # TBD: Take from user input

        # Decompose Unitaries from Data Set into both Gate Set

        samples = len(ds)
        pf01_db, cd01_db = [], []
        pf02_db, cd02_db = [], []
        print("\n  Decomposing Data Set into Gate Set 1:["+gs1_gates+"] and Gate Set 2:["+gs2_gates+"] \n")
        for i in tqdm(range(samples)):   
            pfi, dep, _ = self.dcmp_U_gs(ds[i], gs1)
            pf01_db.append(pfi)
            cd01_db.append(dep)
            pfi, dep, _ = self.dcmp_U_gs(ds[i], gs2)
            pf02_db.append(pfi)
            cd02_db.append(dep)

        return gs1, gs1_gates, pf01_db, cd01_db, gs2, gs2_gates, pf02_db, cd02_db

    # ------------------------------------------------------------------------------------------------ #

    def cnfg_cost_fab(self):

        self.cost_agf = 0
        # Use this method to set fabrication cost of gates in the gate set

        return
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Configure the weights of the elements of the cost function
    """

    def cnfg_wgts(self, wgts):

        self.w_apf = wgts[0]   # Weight of average process fidelity
        self.w_npf = wgts[1]   # Weight of novelty of process fidelity
        self.w_acd = wgts[2]   # Weight of average circuit depth
        self.w_ncd = wgts[3]   # Weight of novelty of circuit depth
        self.w_agf = wgts[4]   # Weight of average gate fabrication
        
        self.cnfg_cost_fab()

        return
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Calculate cost function based on distribution of process fidelity differences and gate depths of two gate sets
    """

    def cfn_calc(self,pf01_db,cd01_db,pf02_db,cd02_db):

        ivt_pf_gs01 = np.subtract(1,pf01_db)
        dist_pf_novelty = np.mean(abs(np.subtract(ivt_pf_gs01,pf02_db)))
        ivt_cd_gs01 = np.subtract(max(cd01_db),cd01_db)
        dist_cd_novelty = np.mean(abs(np.subtract(ivt_cd_gs01,cd02_db)))
        dist_pf_avg = - np.mean(pf02_db)
        dist_cd_avg = np.mean(cd02_db)

        cfn = self.w_apf*dist_pf_avg + self.w_npf*dist_pf_novelty + self.w_acd*dist_cd_avg + self.w_ncd*dist_cd_novelty + self.w_agf*self.cost_agf

        return cfn

    # ------------------------------------------------------------------------------------------------ #
    
    def nusa(self, ds, ngs_cfg, search):

        # Define Gate Set 1 here (TBD: Load saved gate set from file)
        gs1, gs1_gates = self.def_gs(['H1','T1','CX2'])         # TBD: Take from user input

        # Decompose Unitaries from Data Set into Gate Set 1
        samples = len(ds)
        pf01_db, cd01_db = [], []
        print("\n  Decomposing Data Set into Gate Set 1:["+gs1_gates+"] \n")
        for i in tqdm(range(samples)):   
            pf, cd, _ = self.dcmp_U_gs(ds[i], gs1)
            pf01_db.append(pf)
            cd01_db.append(cd)

        # Cost function for optimization
        method = 'COBYLA'
        maxiter = 20
        def cost_to_optimize(gs2_params):
            gs2, _ = self.def_gs(ngs_cfg, gs2_params) 
            pf02_db, cd02_db = [], []
            for i in range(samples):   
                pf, cd, _ = self.dcmp_U_gs(ds[i], gs2)
                pf02_db.append(pf)
                cd02_db.append(cd)
            cfn = self.cfn_calc(pf01_db, cd01_db, pf02_db, cd02_db)
            return cfn

        # Optimize Gate Set 2
        param_ctr = self.gs_param_ctr(ngs_cfg)
        cfn_best, cfn_best_db = np.inf, []
        max_time = 10       # in seconds
        trials = 0
        start = time.time()
        end = time.time()

        while (end - start) < max_time:     # Constraints: time/cost/trials
            trials += 1
            print("\n  Decomposing Data Set into Gate Set 2, trials = "+str(trials)+"\n") 
            params = np.random.rand(param_ctr)
            if search == 1:    # Random search
                gs2, gs2_gates = self.def_gs(ngs_cfg, params) 
                pf02_db, cd02_db = [], []
                for i in range(samples):   
                    pf, cd, _ = self.dcmp_U_gs(ds[i], gs2)
                    pf02_db.append(pf)
                    cd02_db.append(cd)
                cfn = self.cfn_calc(pf01_db, cd01_db, pf02_db, cd02_db)
                if cfn <= cfn_best:
                    cfn_best = cfn
                    cfn_best_db = [gs2, gs2_gates, pf02_db, cd02_db, params] 
            else:               # SciPy optimize
                res = minimize(cost_to_optimize, params, method = method, options={'maxiter': maxiter})
                if res['fun'] <= cfn_best:
                    cfn_best = res['fun']
                    gs2, gs2_gates = self.def_gs(ngs_cfg, res['x'])
                    pf02_db, cd02_db = [], []
                    for i in range(samples):   
                        pf, cd, _ = self.dcmp_U_gs(ds[i], gs2)
                        pf02_db.append(pf)
                        cd02_db.append(cd)
                    cfn_best_db = [gs2, gs2_gates, pf02_db, cd02_db, res['x']] 
            end = time.time()

        return gs1, gs1_gates, pf01_db, cd01_db, cfn_best_db[0], cfn_best_db[1], cfn_best_db[2], cfn_best_db[3], cfn_best_db[4]

    # ------------------------------------------------------------------------------------------------ #


from qiskit import QuantumCircuit
from qiskit.quantum_info import random_statevector, random_unitary, Operator, Statevector
from qiskit.extensions import UnitaryGate
import random
import math
from astropy.coordinates import cartesian_to_spherical
import weylchamber
import numpy as np
from itertools import product
import qutip as qt
from qutip.measurement import measurement_statistics
import matplotlib.pyplot as plt

class GenerateDataSet:

    # ------------------------------------------------------------------------------------------------ #

    def yaqq_gen_ds(self, ds_dim, ds_type, ds_size, ds_reso):

        if ds_dim == 1 and ds_type == 4:
            ds = self.gen_ds_equiA(ds_reso)
            ds_size = len(ds)
            print("\n  ===> YAQQ Data Set Generated for Dimension =", ds_dim, "Type =", ds_type, "Spacing =", ds_reso, "Size =", ds_size)
        elif ds_dim == 2 and ds_type == 4:
            ds = self.gen_ds_equiNL(ds_reso)
            ds_size = len(ds)
            print("\n  ===> YAQQ Data Set Generated for Dimension =", ds_dim, "Type =", ds_type, "Spacing =", ds_reso, "Size =", ds_size)
        else:
            if ds_type == 1:
                ds = self.gen_ds_randS(ds_dim,  ds_size)
            elif ds_type == 2:
                ds = self.gen_ds_randU(max_dim = ds_dim,  ds_size = ds_size, rand_dim = False)
            elif ds_dim == 1 and ds_type == 3:
                ds = self.gen_ds_fiboS(ds_size)
            elif ds_dim == 2 and ds_type == 3:
                ds = self.gen_ds_randNL(ds_size)
            print("\n  ===> YAQQ Data Set Generated for Dimension = "+str(ds_dim)+", Type = "+str(ds_type)+", Size = "+str(ds_size))

        return ds

    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Haar Random n-qubit pure States
    Ref: https://qiskit.org/documentation/stubs/qiskit.quantum_info.random_statevector.html
    """

    def gen_ds_randS(self, ds_dim = 1,  ds_size = 100):

        ds = []
        for i in range(ds_size):
            qc = QuantumCircuit(ds_dim)
            randS = random_statevector(2**ds_dim).data
            qc.prepare_state(randS, list(range(0, ds_dim)))   
            randU_0 = Operator.from_circuit(qc)
            ds.append(UnitaryGate(randU_0,label='RndU'+str(i)))

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Haar Random 2^nx2^n Unitaries
    Ref: https://qiskit.org/documentation/stubs/qiskit.quantum_info.random_unitary.html
    """

    def gen_ds_randU(self, max_dim = 1, ds_size = 100, rand_dim = False):

        ds = []
        for i in range(ds_size):
            dim = max_dim
            if rand_dim:
                dim =  random.randrange(1,max_dim+1)    # TBD: Samples should be proportional to dimension instead of uniform, i.e. exponentially more higher dimension than lower dimensions
            ds.append(UnitaryGate(random_unitary(2**dim),label='RndU'+str(i)))

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Equispaced States on Bloch Sphere using Golden mean
    Ref: https://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
    """

    def gen_ds_fiboS(self, ds_size = 100):

        ds = []
        phi = math.pi * (3. - math.sqrt(5.))        # golden angle in radians
        for i in range(ds_size):
            y = 1 - (i / float(ds_size - 1)) * 2    # y goes from 1 to -1
            radius = math.sqrt(1 - y * y)           # radius at y
            theta = phi * i                         # golden angle increment
            x = math.cos(theta) * radius
            z = math.sin(theta) * radius
            sphe_coor = cartesian_to_spherical(x, y, z)
            z_ang = sphe_coor[1].radian+math.pi/2
            x_ang = sphe_coor[2].radian
            qc = QuantumCircuit(1)       
            qc.ry(z_ang,0)  # To rotate state z_ang from |0>, rotate about Y
            qc.rz(x_ang,0)  # To rotate state x_ang from |+>, rotate about Z
            fiboU = Operator.from_circuit(qc)      
            ds.append(UnitaryGate(fiboU,label='FiboU'+str(i)))

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Equispaced angles on Bloch Sphere
    parameters a_rz and a_rx are re-interpreted in spherical coordinates as respectively the colatitude with respect to the z-axis and the longitude with respect to the x-axis
    """

    def gen_ds_equiA(self, px = 20):

        ds = []
        a_rz = np.linspace(0, math.pi, px, endpoint=False)
        a_rx = np.linspace(0, 2*math.pi, 2*px, endpoint=False)
        ang_ds = product(a_rz, a_rx)
        points = 0
        for ang in ang_ds:
            # Enumerate points in the Bloch sphere
            qc = QuantumCircuit(1)       
            qc.ry(ang[0],0)  # To rotate state z_ang from |0>, rotate about Y
            qc.rz(ang[1],0)  # To rotate state x_ang from |+>, rotate about Z
            points+= 1
            equiU = Operator.from_circuit(qc)      
            ds.append(UnitaryGate(equiU,label='EquiU'+str(points)))

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Random Non-local Unitaries on Weyl chamber
    Ref: https://weylchamber.readthedocs.io/en/latest/API/weylchamber.coordinates.html#weylchamber.coordinates.random_gate
    """

    def gen_ds_randNL(self, ds_size = 100):

        ds = []
        for i in range(ds_size):
            ds.append(UnitaryGate(weylchamber.random_gate(),label='RndU'+str(i)))   

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

    """
    Data Set Generation: Equispaced non-local unitaries
    Ref: https://weylchamber.readthedocs.io/en/latest/API/weylchamber.coordinates.html#weylchamber.coordinates.point_in_weyl_chamber
    """

    def gen_ds_equiNL(self, px = 23):

        ds = []
        cx = np.linspace(0, 1, px)
        cy = np.linspace(0, 0.5, int(px/2))
        cz = np.linspace(0, 0.5, int(px/2))
        gs = product(cx, cy, cz)
        valid_points = 0
        for can in gs:
            # Enumerate points in the Weyl chamber
            c = list(can)
            c1,c2,c3 = c[0],c[1],c[2]
            if weylchamber.point_in_weyl_chamber(c1,c2,c3):
                valid_points+= 1
                ds.append(UnitaryGate(weylchamber.canonical_gate(c1,c2,c3),label='RndU'+str(valid_points)))   

        return ds
    
    # ------------------------------------------------------------------------------------------------ #

class VisualizeDataSet:

    # ------------------------------------------------------------------------------------------------ #

    def rgb_to_hex(self, r, g, b):

        return '#{:02x}{:02x}{:02x}'.format(r, g, b)

    # ------------------------------------------------------------------------------------------------ #

    def vis_ds_Bloch(self, ds):

        b = qt.Bloch()
        b.point_marker = ['o']
        b.point_size = [20]
        samples = len(ds)
        color = []

        for i in range(samples):
            qc = QuantumCircuit(1)
            qc.append(ds[i], [0])
            sv = Statevector(qc).data
            b.add_states(qt.Qobj(sv), kind='point')
            _, _, pX = measurement_statistics(qt.Qobj(sv), qt.sigmax())
            _, _, pY = measurement_statistics(qt.Qobj(sv), qt.sigmay())
            _, _, pZ = measurement_statistics(qt.Qobj(sv), qt.sigmaz())
            color.append(self.rgb_to_hex(int(pX[0]*255),int(pY[0]*255),int(pZ[0]*255)))
            
        b.point_color = color
        b.render()
        plt.show()

        return

    # ------------------------------------------------------------------------------------------------ #

    def vis_ds_Weyl(self, ds):

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        for i in ds:
            c1, c2, c3 = weylchamber.c1c2c3(i.to_matrix())
            if weylchamber.point_in_weyl_chamber(c1,c2,c3):
                ax.scatter(c1, c2, c3, s=2, c=self.rgb_to_hex(255-int(255*(math.cos(c1*2*math.pi)+1)/2),255-int(255*(math.cos(c2*2*math.pi)+1)/2),255-int(255*(math.cos(c3*2*math.pi)+1)/2)))
            
        ax.plot3D([0,0.5],[0,0.5],[0,0.5],linestyle='--',color='black')       # Isotropic exchange Swap-alpha gates
        ax.plot3D([1,0.5],[0,0.5],[0,0.5],linestyle='--',color='black')       # Isotropic exchange Swap-alpha gates 
        ax.plot3D([0.5,0.5],[0.5,0.5],[0,0.5],linestyle='--',color='black')   # Parameterized Swap gates
        ax.plot3D([0,1],[0,0],[0,0],linestyle='--',color='black')             # Ising gates
        ax.plot3D([0,0.5],[0,0.5],[0,0],linestyle='--',color='black')         # XY gates
        ax.plot3D([1,0.5],[0,0.5],[0,0],linestyle='--',color='black')         # XY gates
        # Perfect Entangler boundary
        ax.plot3D([0.25,0.75],[0.25,0.25],[0.25,0.25],linestyle=':',color='grey')
        ax.plot3D([0.25,0.5],[0.25,0],[0.25,0],linestyle=':',color='grey')
        ax.plot3D([0.5,0.75],[0,0.25],[0,0.25],linestyle=':',color='grey')
        ax.plot3D([0.25,0.25],[0.25,0.25],[0.25,0],linestyle=':',color='grey')
        ax.plot3D([0.75,0.75],[0.25,0.25],[0.25,0],linestyle=':',color='grey')
        ax.plot3D([0.5,0.25],[0,0.25],[0,0],linestyle=':',color='grey')
        ax.plot3D([0.5,0.75],[0,0.25],[0,0],linestyle=':',color='grey')
        ax.plot3D([0.25,0.5],[0.25,0.5],[0.25,0],linestyle=':',color='grey')
        ax.plot3D([0.75,0.5],[0.25,0.5],[0.25,0],linestyle=':',color='grey')

        ax.view_init(elev=15, azim=-55)
        tmp_planes = ax.zaxis._PLANES 
        ax.zaxis._PLANES = ( tmp_planes[2], tmp_planes[3], 
                            tmp_planes[0], tmp_planes[1], 
                            tmp_planes[4], tmp_planes[5])
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        ax.set_zlim(0, 1)
        ax.xaxis.set_ticks([0, 0.25, 0.5, 0.75, 1])
        ax.yaxis.set_ticks([0, 0.25, 0.5, 0.75, 1])
        ax.zaxis.set_ticks([0, 0.25, 0.5, 0.75, 1])
        ax.set_xlabel('$c_1/\pi$')
        ax.set_ylabel('$c_2/\pi$')
        ax.zaxis.set_rotate_label(False) 
        ax.set_zlabel('$c_3/\pi$', rotation=90)
        # ax.grid(False)
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False

        plt.show()

        return

    # ------------------------------------------------------------------------------------------------ #

class ResultsPlotSave:

    # ------------------------------------------------------------------------------------------------ #

    def plot_compare_gs(self, gs1, gs1_gates, pf1, cd1, gs2, gs2_gates, pf2, cd2, pfivt = False):
        
        avg_fid_gs01 = np.mean(pf1)
        avg_fid_gs02 = np.mean(pf2)
        avg_dep_gs01 = np.mean(cd1)
        avg_dep_gs02 = np.mean(cd2) 
        
        ivt_fid_gs01 = np.subtract(1,pf1)

        _, ax = plt.subplots(1, 2)
        ax[0].plot(pf1, '-x', color = 'r', label = 'PF ['+gs1_gates+']')
        ax[0].plot(pf2, '-o', color = 'b', label = 'PF ['+gs2_gates+']')
        if pfivt:
            ax[0].plot(ivt_fid_gs01, '-x', color = 'g', label = 'target PF trend')

        ax[0].axhline(y=avg_fid_gs01, linestyle='-.', color = 'r' , label = 'avg.PF ['+gs1_gates+']')
        ax[0].axhline(y=avg_fid_gs02, linestyle='-.', color = 'b' , label = 'avg.PF ['+gs2_gates+']')

        ax[1].plot(cd1, '-x', color = 'r', label = 'CD ['+gs1_gates+']')
        ax[1].plot(cd2, '-o', color = 'b', label = 'CD ['+gs2_gates+']')

        ax[1].axhline(y=avg_dep_gs01, linestyle='-.', color = 'r', label = 'avg.CD ['+gs1_gates+']')
        ax[1].axhline(y=avg_dep_gs02, linestyle='-.', color = 'b', label = 'avg.CD ['+gs2_gates+']')

        ax[0].set_ylabel("Process Fidelity")
        ax[1].set_ylabel("Circuit Depth")
        ax[0].set_ylim(bottom=0,top=1)
        ax[1].set_ylim(bottom=0,top=None)
        ax[0].legend()
        ax[1].legend()
        # _, ax = plt.subplots(1, 2, figsize = (7,3.5), sharex=True, layout="constrained")
        # ax[0].set_xlabel("Equidistant Points")
        # ax[1].set_xlabel("Equidistant Points")
        # plt.legend(ncol = 2, bbox_to_anchor = (1, 1.13))

        save_res = input("Save plots and data? [Y/N] (def.: N): ") or 'N'
        if save_res == 'Y':
            exp_id = input("Enter experiment ID: ") or 'exp_1'
            plt.savefig('results/figures/'+exp_id+'.pdf')
            plt.savefig('results/figures/'+exp_id+'.png')      
            np.save('results/data/'+exp_id+'gs1', gs1)
            np.save('results/data/'+exp_id+'gs2', gs2)

        plt.show()



# Copyright © 2023 Quantum Intelligence Research Group
#
# Distributed under terms of the GNU Affero General Public License.

from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='yaqq',
    version='0.13.4',
    install_requires=["numpy >= 1.23.5", "qiskit >= 0.43.3", "astropy >= 5.3.1", "matplotlib >= 3.7.2", "scipy >= 1.11.1", "tqdm >= 4.65.0", "qutip >= 4.7.2", "scikit-learn >= 1.3.0", "weylchamber >= 0.4.0"],
    long_description=long_description,
    long_description_content_type='text/markdown'
)


from __future__ import annotations
import warnings
import collections
import numpy as np
import qiskit.circuit.library.standard_gates as gates
from qiskit.circuit import Gate
from qiskit.quantum_info.operators.predicates import matrix_equal
from qiskit.synthesis.discrete_basis.gate_sequence import GateSequence
import numpy as np
from qiskit.transpiler.passes.synthesis import SolovayKitaev
from qiskit.quantum_info import process_fidelity, Choi
from astropy.coordinates import cartesian_to_spherical
from qiskit import QuantumCircuit
import matplotlib.pyplot as plt
import math
from qiskit.extensions import UnitaryGate, U3Gate
from qiskit.circuit import Gate
from qiskit.quantum_info import random_unitary
import scipy.linalg as la
from qiskit.circuit.gate import Gate
from tqdm import tqdm
from scipy.optimize import minimize 

from typing import Optional, Union
from qiskit.circuit.controlledgate import ControlledGate

class UGate(Gate):
    U = np.identity(2)
    label = ""
    def __init__(self, label, unitary):
        self.U = unitary
        self.label = label
        """Create new gate."""
        super().__init__(label, 1, [], label=label)
    def inverse(self):
        """Invert this gate."""
        return UdgGate(self.label+'dg', self.U)  # self-inverse
    def __array__(self, dtype=None):
        """Return a numpy.array for the U gate."""
        return self.U
   
class UdgGate(Gate):
    U = np.identity(2)
    label = ""
    def __init__(self, label, unitary):
        self.U = unitary
        self.label = label
        """Create new gate."""
        super().__init__(label, 1, [], label=label)
    def inverse(self):
        """Invert this gate."""
        return UGate(self.label[:-2], self.U)  # self-inverse
    def __array__(self, dtype=None):
        """Return a numpy.array for the Udg gate."""
        return la.inv(self.U)

# from qiskit.synthesis.discrete_basis.generate_basis_approximations import generate_basic_approximations
####################################################################################################
# Updated generalized generate_basic_approximations of qiskit
####################################################################################################

class gen_basis_seq:

    Node = collections.namedtuple("Node", ("labels", "sequence", "children"))
    _1q_gates = {}

    def _check_candidate_kdtree(self, candidate, existing_sequences, tol=1e-10):
        from sklearn.neighbors import KDTree
        # do a quick, string-based check if the same sequence already exists
        if any(candidate.name == existing.name for existing in existing_sequences):
            return False
        points = np.array([sequence.product.flatten() for sequence in existing_sequences])
        candidate = np.array([candidate.product.flatten()])
        kdtree = KDTree(points)
        dist, _ = kdtree.query(candidate)
        return dist[0][0] > tol

    def _process_node(self, node: self.Node, basis: list[str], sequences: list[GateSequence]):  
        for label in basis:
            sequence = node.sequence.copy()
            sequence.append(self._1q_gates[label])
            if self._check_candidate_kdtree(sequence, sequences):
                sequences.append(sequence)
                node.children.append(self.Node(node.labels + (label,), sequence, []))
        return node.children

    def generate_basic_approximations(self, basis_gates: list[Gate], depth: int = 3) -> list[GateSequence]:
        """Generates a list of ``GateSequence``s with the gates in ``basic_gates``.
        Args:
            basis_gates: The gates from which to create the sequences of gates.
            depth: The maximum depth of the approximations.
            filename: If provided, the basic approximations are stored in this file.
        Returns:
            List of ``GateSequences`` using the gates in ``basic_gates``.
        """
        
        basis = []
        for gate in basis_gates:
            basis.append(gate.name)
            self._1q_gates[gate.label] = gate

        tree = self.Node((), GateSequence(), [])
        cur_level = [tree]
        sequences = [tree.sequence]
        for _ in [None] * depth:
            next_level = []
            for node in cur_level:
                next_level.extend(self._process_node(node, basis, sequences))
            cur_level = next_level
        return sequences